package tetris;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.Arrays;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

/**
 * ˹Ϸ
 * @author Leslie Leung
 */
public class TetrisPane extends JPanel {
	public static final int ROWS = 20;	//
	public static final int COLUMNS = 16;	//
	
	/* ʾ7ֲͬĸ񷽿 */
	public static final int I_SHAPED = 0;
	public static final int S_SHAPED = 1;
	public static final int T_SHAPED = 2;
	public static final int Z_SHAPED = 3;
	public static final int L_SHAPED = 4;
	public static final int O_SHAPED = 5;
	public static final int J_SHAPED = 6;
	
	public static final int KIND = 7;	//ʾĸ񷽿7
	public static final int INIT_SPEED = 1000;	//ʾĳʼٶ
	
	private static int randomNum = 0;	//ʾɵĶ˹Ŀ
	
	private Random random;
	private Tetromino currentTetromino;	//ʾǰĸ񷽿
	private Cell[][] wall;		//ʾǽnullʾû
	private Timer autoDrop;		//ʵԶļʱ
	private KeyControl keyListener;	//ʾ¼ر
	
	/**
	 * 췽
	 */
	public TetrisPane() {
		setPreferredSize(new Dimension(COLUMNS * Cell.CELL_SIZE, ROWS * Cell.CELL_SIZE));
		
		random = new Random();
		wall = new Cell[ROWS][COLUMNS];
		autoDrop = new Timer();
		keyListener = new KeyControl();
		
		randomOne();
		
		autoDrop.schedule(new DropExecution(), (long)interval(), (long)interval());
	}
	
	/**
	 * һĸ񷽿
	 */
	public void randomOne() {
		Tetromino tetromino = null;
		
		/* 7ĸ񷽿һ */
		switch(random.nextInt(KIND)) {
			case I_SHAPED: 
				tetromino = new IShaped();
				break;
			case S_SHAPED: 
				tetromino = new SShaped();
			   	break;
			case T_SHAPED: 
				tetromino = new TShaped();
			    break;
			case Z_SHAPED: 
				tetromino = new ZShaped();
			    break;
			case L_SHAPED: 
				tetromino = new LShaped();
			    break;
			case O_SHAPED: 
				tetromino = new OShaped();
			    break;
			case J_SHAPED: 
				tetromino = new JShaped();
			    break;
		}
		currentTetromino = tetromino;	//ǰĸ񷽿Ϊɵĸ񷽿
		randomNum ++;
	}
	
	/**
	 * жǷ
	 * @return trueˣfalseû
	 */
	public boolean isGameOver() {
		int x, y;	//ǰ˹ӵĺ
		for(int i = 0; i < getCurrentCells().length; i ++) {
			x = getCurrentCells()[i].getX();
			y = getCurrentCells()[i].getY();
			
			if(isContain(x, y)) {//ɵλǷѴڷ󣬴ڵĻʾ
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * ÿһ˹飬ͨıTimerTaskʱӿٶ
	 * @return ʱ
	 */
	public double interval() {
		return INIT_SPEED * Math.pow((double)39 / 38, 0 - randomNum);
	}
	
	/**
	 * KeyControlʵ
	 * @return KeyControlʵ
	 */
	public KeyControl getInnerInstanceOfKeyControl() {
		return keyListener;
	}
	
	/**
	 * ڲ࣬ʵֶ˹Զ
	 * @author Leslie Leung
	 */
	private class DropExecution extends TimerTask {	
		@Override
		public void run() {
			// TODO Auto-generated method stub
			
			if(isGameOver()) {//
				JOptionPane.showMessageDialog(null, "");
				autoDrop.cancel();
				removeKeyListener(keyListener);
				return;
			}
			
			if(!isReachBottomEdge()) {
				currentTetromino.softDrop();
			} else {
				landIntoWall();		//Ѷ˹ӵǽ
				removeRows();	//Уɾ
				randomOne();	//½һ˹
				
				autoDrop.cancel();
				autoDrop = new Timer();
				autoDrop.schedule(new DropExecution(), (long)interval(), (long)interval());
			}
			
			repaint();
		}	
	}
	
	/**
	 * Ѷ˹ӵǽ
	 */
	public void landIntoWall() {
		int x, y;	//˹鲻ƶĺ
		
		for(int i = 0; i < getCurrentCells().length; i ++) {
			x = getCurrentCells()[i].getX();
			y = getCurrentCells()[i].getY();
			
			wall[y][x] = getCurrentCells()[i];	//ӵǽ
		}
	}
	
	/**
	 * ڲ࣬ʵ̵ּ¼
	 * @author Leslie Leung
	 */
	private class KeyControl extends KeyAdapter {
		@Override
		public void keyPressed(KeyEvent e) {
			switch(e.getKeyCode()) {
				case KeyEvent.VK_LEFT: //ƶ
					
					if(!isReachLeftEdge()) {//˹û߽ʱƶ
						currentTetromino.moveLeft();
						repaint();
					}					
					break;
					
				case KeyEvent.VK_RIGHT:	//ƶ
					
					if(!isReachRightEdge()) {//˹ûұ߽ʱƶ
						currentTetromino.moveRight();
						repaint();
					}
					break;
				
				case KeyEvent.VK_DOWN:	//ƶ
					
					if(!isReachBottomEdge()) {//˹û±߽ʱƶ
						currentTetromino.softDrop();
						repaint();
					}
					
					break;
					
				case KeyEvent.VK_SPACE:	//Ӳ
					
					hardDrop();	//Ӳ
					landIntoWall();		//ӵǽ
					removeRows();	//Уɾ
					
					randomOne();
					autoDrop.cancel();
					autoDrop = new Timer();
					autoDrop.schedule(new DropExecution(), (long)interval(), (long)interval());
					
					repaint();
					break;
					
				case KeyEvent.VK_D:	//˳ʱת
					
					if(!clockwiseRotateIsOutOfBounds() && !(currentTetromino instanceof OShaped)) {//˹ûԽ䲻ΪOʱת
						currentTetromino.clockwiseRotate(getAxis(), getRotateCells());
						repaint();
					}
					break;
					
				case KeyEvent.VK_A:	//ʱת
					
					if(!anticlockwiseRotateIsOutOfBounds() && !(currentTetromino instanceof OShaped)) {//˹ûԽ䲻ΪOʱת
						currentTetromino.anticlockwiseRotate(getAxis(), getRotateCells());
						repaint();
					}
					break;
			}
		}
	}
	
	/**
	 * ڲ࣬Iεĸ񷽿飬̳Tetromino
	 * @author Leslie Leung
	 */
	private class IShaped extends Tetromino {
		/**
		 * 췽
		 */
		public IShaped() {
			cells = new Cell[4];
			
			cells[1] = new Cell(3, 0, Cell.COLOR_CYAN);
			cells[0] = new Cell(4, 0, Cell.COLOR_CYAN);
			cells[2] = new Cell(5, 0, Cell.COLOR_CYAN);
			cells[3] = new Cell(6, 0, Cell.COLOR_CYAN);
			
			/* תҪתĸ */
			setAxis();
			setRotateCells();
			
			repaint();
		}
	}
	
	/**
	 * ڲ࣬Sεĸ񷽿飬̳Tetromino
	 * @author Leslie Leung
	 */
	private class SShaped extends Tetromino {
		/**
		 * 췽
		 */
		public SShaped() {
			cells = new Cell[4];
			
			cells[0] = new Cell(4, 0, Cell.COLOR_BLUE);
			cells[1] = new Cell(5, 0, Cell.COLOR_BLUE);
			cells[2] = new Cell(3, 1, Cell.COLOR_BLUE);
			cells[3] = new Cell(4, 1, Cell.COLOR_BLUE);
			
			/* תҪתĸ */
			setAxis();
			setRotateCells();
			
			repaint();
		}
	}
	
	/**
	 * ڲ࣬Tεĸ񷽿飬̳Tetromino
	 * @author Leslie Leung
	 */
	private class TShaped extends Tetromino {
		/**
		 * 췽
		 */
		public TShaped() {
			cells = new Cell[4];
			
			cells[1] = new Cell(3, 0, Cell.COLOR_GREEN);
			cells[0] = new Cell(4, 0, Cell.COLOR_GREEN);
			cells[2] = new Cell(5, 0, Cell.COLOR_GREEN);
			cells[3] = new Cell(4, 1, Cell.COLOR_GREEN);
			
			/* תҪתĸ */
			setAxis();
			setRotateCells();
			
			repaint();
		}
	}
	
	/**
	 * ڲ࣬Zεĸ񷽿飬̳Tetromino
	 * @author Leslie Leung
	 */
	private class ZShaped extends Tetromino {
		/**
		 * 췽
		 */
		public ZShaped() {
			cells = new Cell[4];
			
			cells[1] = new Cell(3, 0, Cell.COLOR_ORANGE);
			cells[2] = new Cell(4, 0, Cell.COLOR_ORANGE);
			cells[0] = new Cell(4, 1, Cell.COLOR_ORANGE);
			cells[3] = new Cell(5, 1, Cell.COLOR_ORANGE);
			
			/* תҪתĸ */
			setAxis();
			setRotateCells();
			
			repaint();
		}
	}
	
	/**
	 * ڲ࣬Lεĸ񷽿飬̳Tetromino
	 * @author Leslie Leung
	 */
	private class LShaped extends Tetromino {
		/**
		 * 췽
		 */
		public LShaped() {
			cells = new Cell[4];
			
			cells[1] = new Cell(3, 0, Cell.COLOR_PINK);
			cells[0] = new Cell(4, 0, Cell.COLOR_PINK);
			cells[2] = new Cell(5, 0, Cell.COLOR_PINK);
			cells[3] = new Cell(3, 1, Cell.COLOR_PINK);
			
			/* תҪתĸ */
			setAxis();
			setRotateCells();
			
			repaint();
		}
	}
	
	/**
	 * ڲ࣬Oεĸ񷽿飬̳Tetromino
	 * @author Leslie Leung
	 */
	private class OShaped extends Tetromino {
		/**
		 * 췽
		 */
		public OShaped() {
			cells = new Cell[4];
			
			cells[0] = new Cell(4, 0, Cell.COLOR_RED);
			cells[1] = new Cell(5, 0, Cell.COLOR_RED);
			cells[2] = new Cell(4, 1, Cell.COLOR_RED);
			cells[3] = new Cell(5, 1, Cell.COLOR_RED);
			
			/* תҪתĸ */
			setAxis();
			setRotateCells();
			
			repaint();
		}
	}
	
	/**
	 * ڲ࣬Jεĸ񷽿飬̳Tetromino
	 * @author Leslie Leung
	 */
	private class JShaped extends Tetromino {
		/**
		 * 췽
		 */
		public JShaped() {
			cells = new Cell[4];
			
			cells[1] = new Cell(3, 0, Cell.COLOR_YELLOW);
			cells[0] = new Cell(4, 0, Cell.COLOR_YELLOW);
			cells[2] = new Cell(5, 0, Cell.COLOR_YELLOW);
			cells[3] = new Cell(5, 1, Cell.COLOR_YELLOW);
			
			/* תҪתĸ */
			setAxis();
			setRotateCells();
			
			repaint();
		}
	}
	
	/**
	 * ɾ
	 */
	public void removeRows() {
		for(int i = 0; i < getCurrentCells().length; i ++) {
			removeRow(getCurrentCells()[i].getY());
		}
	}
	
	/**
	 * ȡת
	 * @return ת
	 */
	public Cell getAxis() {
		return currentTetromino.getAxis();
	}
	
	/**
	 * ȡҪתĸ
	 * @return Ҫתĸ
	 */
	public Cell[] getRotateCells() {
		return currentTetromino.getRotateCells();
	}
	
	/**
	 * ȡǰ˹и
	 * @return ǰ˹и
	 */
	public Cell[] getCurrentCells() {
		return currentTetromino.getCells();
	}
	
	/**
	 * ж϶˹Ƿ񵽴ײ(Ƿ񵽴ײһλǷз)
	 * @return truefalseû
	 */
	public boolean isReachBottomEdge() {
		int oldY, newY, oldX;		//Ӿɵꡢµ;ɵĺ
		
		for(int i = 0; i < getCurrentCells().length; i ++) {
			oldY = getCurrentCells()[i].getY();
			newY = oldY + 1;
			oldX = getCurrentCells()[i].getX();
			
			if(oldY == ROWS - 1) {//ײ
				return true;
			}
			
			if(isContain(oldX, newY)) {//һλз
				return true;
			}
		}
		return false;
	}
	
	/**
	 * ж϶˹Ƿ񵽴߽(Ƿ񳬳߽һλǷײ)
	 * @return trueѵfalseû
	 */
	public boolean isReachLeftEdge() {
		int oldX, newX, oldY;		//Ӿɵĺꡢµĺ;ɵ
		
		for(int i = 0; i < getCurrentCells().length; i ++) {
			oldX = getCurrentCells()[i].getX();
			newX = oldX - 1;
			oldY = getCurrentCells()[i].getY();
			
			if(oldX == 0 || isContain(newX, oldY)) {//߽
				return true;
			}
			
			if(isContain(newX, oldY)) {//һλз
				return true;
			}
		}
		return false;
	}
	
	/**
	 * ж϶˹Ƿ񵽴ұ߽(Ƿ񳬳ұ߽һλǷײ)
	 * @return trueѵfalseû
	 */
	public boolean isReachRightEdge() {
		int oldX, newX, oldY;		//Ӿɵĺꡢµĺ;ɵ
		
		for(int i = 0; i < getCurrentCells().length; i ++) {
			oldX = getCurrentCells()[i].getX();
			newX = oldX + 1;
			oldY = getCurrentCells()[i].getY();
			
			if(oldX == COLUMNS - 1 || isContain(newX, oldY)) {//ұ߽
				return true;
			}
			
			if(isContain(newX, oldY)) {//һλз
				return true;
			}
		}
		return false;
	}
	
	/**
	 * ж϶˹˳ʱתǷ񳬳߽(Ƿ񳬳߽һλǷз)
	 * @return true߽磻falseû߽
	 */
	public boolean clockwiseRotateIsOutOfBounds() {
		int oldX;	//rotateCellĺ
		int oldY;	//rotateCell
		int newX;	//rotateCellתĺ
		int newY;	//rotateCellת
		
		for(int i = 0; i < 3; i ++) {
			oldX = getRotateCells()[i].getX();
			oldY = getRotateCells()[i].getY();
			
			newX = getAxis().getX() - oldY + getAxis().getY();	//º㷨
			newY = getAxis().getY() + oldX - getAxis().getX();	//㷨
			
			if(newX < 0 || newY < 0 || newX > COLUMNS - 1 || newY > ROWS - 1) {//Խ磬true
				return true;
			}
			
			if(isContain(newX, newY)) {//Խ磬true
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * ж϶˹ʱתǷ񳬳߽(Ƿ񳬳߽һλǷз)
	 * @return true߽磻falseû߽
	 */
	public boolean anticlockwiseRotateIsOutOfBounds() {
		int oldX;	//rotateCellĺ
		int oldY;	//rotateCell
		int newX;	//rotateCellתĺ
		int newY;	//rotateCellת
		
		for(int i = 0; i < 3; i ++) {
			oldX = getRotateCells()[i].getX();
			oldY = getRotateCells()[i].getY();
			
			newX = getAxis().getX() - getAxis().getY() + oldY;	//º㷨
			newY = getAxis().getY() + getAxis().getX() - oldX;	//㷨
			
			if(newX < 0 || newY < 0 || newX > COLUMNS - 1 || newY > ROWS - 1) {//Խ磬true
				return true;
			}
			
			if(isContain(newX, newY)) {//Խ磬true
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * жĳǷڷ
	 * @param x 
	 * @param y 
	 * @return trueڶfalseڶ
	 */
	public boolean isContain(int x, int y) {
		if(wall[y][x] == null) {
			return false;
		} else {
			return true;
		}
	}
	
	/**
	 * ʵֶ˹Ӳ
	 */
	public void hardDrop() {
		while(!isReachBottomEdge()) {
			currentTetromino.softDrop();
		}
	}
	
	/**
	 * 
	 * @param i е±
	 */
	public void removeRow(int i) {
		int oldY, newY;	
		
		for(int j = 0; j < COLUMNS; j ++) {
			if(wall[i][j] == null) {//һûreturn
				return;
			}
		}
		
		/* вѸķ */
		for(int k = i; k >= 1; k --){
			System.arraycopy(wall[k - 1], 0, wall[k], 0, COLUMNS);
			
			for(int m = 0; m < COLUMNS; m ++) {
				if(wall[k][m] != null) {//ڲǿյĶҪ
					oldY = wall[k][m].getY();
					newY = oldY + 1;
					wall[k][m].setY(newY);				
				}
			}
			
		}
		Arrays.fill(wall[0], null);
	}
	
	@Override
	public void paint(Graphics g) {
		g.setColor(Color.BLACK);
		g.fillRect(0, 0, getBounds().width, getBounds().height);
		
		/* ǽ */
		for(int i = 0; i < ROWS; i ++) {
			for(int j = 0; j < COLUMNS; j ++) {
				if(wall[i][j] == null) {//ĳķΪʱ
					g.setColor(Color.WHITE);
					g.fillRect(j * Cell.CELL_SIZE + 1, i * Cell.CELL_SIZE + 1, Cell.CELL_SIZE - 2, Cell.CELL_SIZE - 2);
				} else {//鲻Ϊʱ
					wall[i][j].paintCell(g);
				}
			}
		}
		
		/* ǰ˹ */
		for(int i = 0; i < getCurrentCells().length; i ++) {
			getCurrentCells()[i].paintCell(g);
		}
		
	}
}
